"use strict";
// Copyright (c) Microsoft Corporation.
// Licensed under the MIT license.
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const base_1 = require("@jupyter-widgets/base");
const powerbi_client_1 = require("powerbi-client");
const version_1 = require("./version");
// Import the CSS
require("../css/report.css");
const utils_1 = require("./utils");
// Initialize powerbi service
const powerbi = new powerbi_client_1.service.Service(powerbi_client_1.factories.hpmFactory, powerbi_client_1.factories.wpmpFactory, powerbi_client_1.factories.routerFactory);
const EXPORT_DATA_DEFAULT_STATE = {
    pageName: undefined,
    visualName: undefined,
    rows: undefined,
    exportDataType: undefined,
};
const REPORT_FILTER_REQUEST_DEFAULT_STATE = {
    filters: [],
    request_completed: true,
};
const REPORT_NOT_EMBEDDED_MESSAGE = 'Power BI report is not embedded';
// Set threshold to refresh token in minutes
const TOKEN_REFRESH_THRESHOLD = 10;
class ReportModel extends base_1.DOMWidgetModel {
    defaults() {
        return Object.assign(Object.assign({}, super.defaults()), { _model_name: ReportModel.model_name, _model_module: ReportModel.model_module, _model_module_version: ReportModel.model_module_version, _view_name: ReportModel.view_name, _view_module: ReportModel.view_module, _view_module_version: ReportModel.view_module_version, _embed_config: {}, _embedded: false, container_height: 0, container_width: 0, _export_visual_data_request: EXPORT_DATA_DEFAULT_STATE, _visual_data: null, _event_data: {
                event_name: null,
                event_details: null,
            }, _get_filters_request: false, _report_filters: [], _report_filters_request: REPORT_FILTER_REQUEST_DEFAULT_STATE, _get_pages_request: false, _report_pages: [], _get_visuals_page_name: null, _page_visuals: [], _report_bookmark_name: null, _get_bookmarks_request: false, _report_bookmarks: [], _token_expired: false, _client_error: null });
    }
}
exports.ReportModel = ReportModel;
ReportModel.serializers = Object.assign({}, base_1.DOMWidgetModel.serializers);
ReportModel.model_name = 'ReportModel';
ReportModel.model_module = version_1.MODULE_NAME;
ReportModel.model_module_version = version_1.MODULE_VERSION;
ReportModel.view_name = 'ReportView';
ReportModel.view_module = version_1.MODULE_NAME;
ReportModel.view_module_version = version_1.MODULE_VERSION;
class ReportView extends base_1.DOMWidgetView {
    render() {
        const newDivElement = document.createElement('div');
        newDivElement.style.visibility = 'hidden';
        this.el.appendChild(newDivElement);
        this.reportContainer = newDivElement;
        this.embedConfigChanged();
        // Observe changes in the traitlets in Python, and define custom callback.
        this.model.on('change:_embed_config', this.embedConfigChanged, this);
        this.model.on('change:container_height', this.dimensionsChanged, this);
        this.model.on('change:container_width', this.dimensionsChanged, this);
        this.model.on('change:_export_visual_data_request', this.exportVisualDataRequestChanged, this);
        this.model.on('change:_get_filters_request', this.getFiltersRequestChanged, this);
        this.model.on('change:_report_filters_request', this.reportFiltersChanged, this);
        this.model.on('change:_get_pages_request', this.getPagesRequestChanged, this);
        this.model.on('change:_get_visuals_page_name', this.getVisualsPageNameChanged, this);
        this.model.on('change:_report_bookmark_name', this.reportBookmarkNameChanged, this);
        this.model.on('change:_get_bookmarks_request', this.getBookmarksRequestChanged, this);
    }
    dimensionsChanged() {
        this.reportContainer.style.height = `${this.model.get('container_height')}px`;
        this.reportContainer.style.width = `${this.model.get('container_width')}px`;
    }
    setTokenExpiredFlag() {
        this.model.set('_token_expired', true);
        this.touch();
    }
    embedConfigChanged() {
        const embedConfig = this.model.get('_embed_config');
        const reportConfig = embedConfig;
        if (this.report) {
            const accessToken = reportConfig.accessToken;
            // To avoid re-embedding the report
            if (this.report.config.embedUrl === reportConfig.embedUrl &&
                this.report.config.accessToken !== reportConfig.accessToken) {
                // Set new access token
                this.report.setAccessToken(accessToken);
                if (embedConfig.tokenExpiration) {
                    // Set token expiration listener to update the token TOKEN_REFRESH_THRESHOLD minutes before expiration
                    utils_1.setTokenExpirationListener(embedConfig.tokenExpiration, TOKEN_REFRESH_THRESHOLD, this);
                }
            }
            this.model.set('_embedded', true);
            this.touch();
            return;
        }
        // Flag for checking create mode
        const createReportMode = embedConfig.viewMode !== powerbi_client_1.models.ViewMode.View &&
            embedConfig.viewMode !== powerbi_client_1.models.ViewMode.Edit;
        // Check if the embedding mode is create mode
        if (createReportMode) {
            // Create blank Power BI report
            this.report = powerbi.createReport(this.reportContainer, reportConfig);
        }
        else {
            // Phased loading
            this.report = powerbi.load(this.reportContainer, reportConfig);
        }
        // Set default aspect ratio
        let aspectRatio = 9 / 16;
        this.report.on('loaded', () => __awaiter(this, void 0, void 0, function* () {
            console.log('Loaded');
            if (embedConfig.tokenExpiration) {
                // Set token expiration listener to update the token TOKEN_REFRESH_THRESHOLD minutes before expiration
                utils_1.setTokenExpirationListener(embedConfig.tokenExpiration, TOKEN_REFRESH_THRESHOLD, this);
            }
            try {
                // Check if embed mode is view/edit to get active page size
                if (!createReportMode) {
                    // Get page size for the Power BI report
                    const { width, height } = yield utils_1.getActivePageSize(this.report);
                    if (width && height) {
                        // Update aspect ratio according to the page size
                        aspectRatio = height / width;
                    }
                    else {
                        console.error('Invalid report size');
                    }
                }
                // Get dimensions of output cell
                const DOMRect = this.el.getBoundingClientRect();
                const outputCellWidth = DOMRect.width || 980;
                const newHeight = outputCellWidth * aspectRatio;
                // Set dimensions of report container
                this.reportContainer.style.width = `${outputCellWidth}px`;
                this.reportContainer.style.height = `${newHeight}px`;
                // Show the report container
                this.reportContainer.style.visibility = 'visible';
                if (!createReportMode) {
                    // Complete the phased embedding
                    this.report.render();
                }
            }
            catch (error) {
                console.error(error);
            }
            // Invoke loaded event handler on kernel side
            this.model.set('_event_data', {
                event_name: 'loaded',
                event_details: null,
            });
            this.touch();
        }));
        this.report.on('rendered', () => {
            console.log('Rendered');
            // Invoke rendered event handler on kernel side
            this.model.set('_event_data', {
                event_name: 'rendered',
                event_details: null,
            });
            this.touch();
        });
        this.report.on('error', (errorMessage) => {
            // Invoke error event handler on kernel side
            this.model.set('_event_data', {
                event_name: 'error',
                event_details: errorMessage.detail,
            });
            this.touch();
        });
        // Notify that report embedding has started
        this.model.set('_embedded', true);
        // Commit model changes
        this.touch();
    }
    exportVisualDataRequestChanged() {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.report) {
                console.error(REPORT_NOT_EMBEDDED_MESSAGE);
                this.model.set('_client_error', REPORT_NOT_EMBEDDED_MESSAGE);
                this.touch();
                return;
            }
            const exportVisualDataRequest = this.model.get('_export_visual_data_request');
            // Check export visual data request object is null or empty
            if (!exportVisualDataRequest.pageName &&
                !exportVisualDataRequest.visualName &&
                !exportVisualDataRequest.rows &&
                !exportVisualDataRequest.exportDataType) {
                // This is the case of model reset
                return;
            }
            if (!exportVisualDataRequest.pageName || !exportVisualDataRequest.visualName) {
                console.error('Page and visual names are required');
                this.model.set('_client_error', 'Page and visual names are required');
                this.touch();
                return;
            }
            const pageName = exportVisualDataRequest.pageName;
            const visualName = exportVisualDataRequest.visualName;
            const dataRows = exportVisualDataRequest.rows;
            const exportDataType = exportVisualDataRequest.exportDataType;
            try {
                const selectedPage = yield utils_1.getRequestedPage(this.report, pageName);
                const visuals = yield selectedPage.getVisuals();
                const selectedVisual = visuals.filter((visual) => {
                    return visual.name === visualName;
                })[0];
                if (!selectedVisual) {
                    throw 'Visual not found';
                }
                const data = yield selectedVisual.exportData(exportDataType, dataRows);
                // Update data
                this.model.set('_visual_data', data.data);
                this.touch();
            }
            catch (error) {
                let stringifiedError = JSON.stringify(error);
                if (stringifiedError === '{}') {
                    stringifiedError = error.toString();
                }
                console.error(error);
                this.model.set('_client_error', stringifiedError);
                this.touch();
            }
        });
    }
    getFiltersRequestChanged() {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.report) {
                console.error(REPORT_NOT_EMBEDDED_MESSAGE);
                this.model.set('_client_error', REPORT_NOT_EMBEDDED_MESSAGE);
                this.touch();
                return;
            }
            const get_filters_request = this.model.get('_get_filters_request');
            if (!get_filters_request) {
                // Reset of get filters request
                return;
            }
            try {
                // Get list of filters applied on the report
                const filters = yield this.report.getFilters();
                if (!filters) {
                    throw 'No filters available';
                }
                this.model.set('_report_filters', filters);
                this.touch();
            }
            catch (error) {
                let stringifiedError = JSON.stringify(error);
                if (stringifiedError === '{}') {
                    stringifiedError = error.toString();
                }
                console.error(error);
                this.model.set('_client_error', stringifiedError);
                this.touch();
            }
        });
    }
    reportFiltersChanged() {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.report) {
                console.error(REPORT_NOT_EMBEDDED_MESSAGE);
                this.model.set('_client_error', REPORT_NOT_EMBEDDED_MESSAGE);
                this.touch();
                return;
            }
            const filterRequest = this.model.get('_report_filters_request');
            if (filterRequest.request_completed) {
                // Reset of filter request
                return;
            }
            try {
                // Add new filters or remove filters when filters array is empty
                if (filterRequest.filters.length > 0) {
                    yield this.report.updateFilters(powerbi_client_1.models.FiltersOperations.Replace, filterRequest.filters);
                }
                else {
                    yield this.report.updateFilters(powerbi_client_1.models.FiltersOperations.RemoveAll);
                }
                // Reset filter request
                this.model.set('_report_filters_request', REPORT_FILTER_REQUEST_DEFAULT_STATE);
                this.touch();
            }
            catch (error) {
                let stringifiedError = JSON.stringify(error);
                if (stringifiedError === '{}') {
                    stringifiedError = error.toString();
                }
                console.error(error);
                this.model.set('_client_error', stringifiedError);
                this.touch();
            }
        });
    }
    getPagesRequestChanged() {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.report) {
                console.error(REPORT_NOT_EMBEDDED_MESSAGE);
                this.model.set('_client_error', REPORT_NOT_EMBEDDED_MESSAGE);
                this.touch();
                return;
            }
            const get_pages_request = this.model.get('_get_pages_request');
            if (!get_pages_request) {
                return;
            }
            try {
                const pages = yield this.report.getPages();
                if (!pages) {
                    throw 'Pages not found';
                }
                // Remove 'report' property from Page object to handle nested property loop
                const pagesWithoutReport = pages.map((page) => {
                    delete page.report;
                    return page;
                });
                this.model.set('_report_pages', pagesWithoutReport);
                this.touch();
            }
            catch (error) {
                let stringifiedError = JSON.stringify(error);
                if (stringifiedError === '{}') {
                    stringifiedError = error.toString();
                }
                console.error(error);
                this.model.set('_client_error', stringifiedError);
                this.touch();
            }
        });
    }
    getVisualsPageNameChanged() {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.report) {
                console.error(REPORT_NOT_EMBEDDED_MESSAGE);
                this.model.set('_client_error', REPORT_NOT_EMBEDDED_MESSAGE);
                this.touch();
                return;
            }
            const get_visuals_page_name = this.model.get('_get_visuals_page_name');
            if (!get_visuals_page_name) {
                return;
            }
            try {
                const selectedPage = yield utils_1.getRequestedPage(this.report, get_visuals_page_name);
                const visuals = yield selectedPage.getVisuals();
                if (!visuals) {
                    throw 'Visuals not found';
                }
                // Remove 'page' property from Visual object to handle nested property loop
                const visualsWithoutPage = visuals.map((visual) => {
                    delete visual.page;
                    return visual;
                });
                this.model.set('_page_visuals', visualsWithoutPage);
                this.touch();
            }
            catch (error) {
                let stringifiedError = JSON.stringify(error);
                if (stringifiedError === '{}') {
                    stringifiedError = error.toString();
                }
                console.error(error);
                this.model.set('_client_error', stringifiedError);
                this.touch();
            }
        });
    }
    reportBookmarkNameChanged() {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.report) {
                console.error(REPORT_NOT_EMBEDDED_MESSAGE);
                return;
            }
            const bookmarkName = this.model.get('_report_bookmark_name');
            try {
                // Apply corresponding bookmark to the embedded report
                yield this.report.bookmarksManager.apply(bookmarkName);
            }
            catch (error) {
                console.error('Set bookmark error:', error);
            }
        });
    }
    getBookmarksRequestChanged() {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.report) {
                console.error(REPORT_NOT_EMBEDDED_MESSAGE);
                this.model.set('_client_error', REPORT_NOT_EMBEDDED_MESSAGE);
                this.touch();
                return;
            }
            const get_bookmarks_request = this.model.get('_get_bookmarks_request');
            if (!get_bookmarks_request) {
                return;
            }
            try {
                // Get list of bookmarks present in the report
                const bookmarks = yield this.report.bookmarksManager.getBookmarks();
                // Check if there is any bookmark saved on the embedded report
                if (bookmarks.length === 0) {
                    this.model.set('_report_bookmarks', ['']);
                    this.touch();
                }
                else {
                    this.model.set('_report_bookmarks', bookmarks);
                    this.touch();
                }
            }
            catch (error) {
                let stringifiedError = JSON.stringify(error);
                if (stringifiedError === '{}') {
                    stringifiedError = error.toString();
                }
                console.error(error);
                this.model.set('_client_error', stringifiedError);
                this.touch();
            }
        });
    }
}
exports.ReportView = ReportView;
//# sourceMappingURL=report.js.map